import json as jsonlib
import os
import typing
from json import JSONEncoder

from starlette.datastructures import URL, URLPath
from starlette.requests import Request
from starlette.responses import (
    ContentStream,
    FileResponse,
    HTMLResponse,
    JSONResponse as JSONResponse_,
    PlainTextResponse,
    RedirectResponse,
    Response,
    StreamingResponse,
)

from kupala.templating import Templates


class JSONResponse(JSONResponse_):
    def __init__(
        self,
        content: typing.Any,
        status_code: int = 200,
        headers: typing.Mapping[str, str] | None = None,
        media_type: str | None = None,
        indent: int = 2,
        encoder_class: type[JSONEncoder] | None = None,
    ):
        self._indent = indent
        self._encoder_class = encoder_class
        super().__init__(content, status_code, headers, media_type)

    def render(self, content: typing.Any) -> bytes:
        return jsonlib.dumps(
            content,
            ensure_ascii=False,
            allow_nan=False,
            indent=self._indent,
            separators=(",", ":"),
            cls=self._encoder_class,
        ).encode("utf-8")


class EmptyResponse(Response):
    def __init__(self, headers: typing.Mapping[str, str] | None = None) -> None:
        super().__init__(b"", status_code=204, headers=headers)


class GoBackResponse(RedirectResponse):
    def __init__(
        self,
        request: Request,
        status_code: int = 302,
        headers: typing.Mapping[str, str] | None = None,
    ) -> None:
        default_url = URLPath(str(request.base_url), request.url.scheme, request.url.netloc)
        redirect_to = request.headers.get("referer", default_url)
        current_origin = request.url.netloc
        if current_origin not in redirect_to:
            redirect_to = default_url

        super().__init__(
            redirect_to,
            status_code=status_code,
            headers=headers,
        )


class JSONErrorResponse(JSONResponse):
    def __init__(
        self,
        message: str,
        errors: dict[str, list[str]] | None = None,
        status_code: int = 200,
        headers: typing.Mapping[str, str] | None = None,
        media_type: str | None = None,
    ) -> None:
        content: dict[str, str | int | dict[str, list[str]]] = {
            "message": message,
            "errors": errors or {},
        }
        super().__init__(content, status_code, headers, media_type)


def text(content: str, status_code: int = 200, headers: typing.Mapping[str, str] | None = None) -> PlainTextResponse:
    """Return a plain text response."""
    return PlainTextResponse(content, status_code, headers=headers)


def html(content: str, status_code: int = 200, headers: typing.Mapping[str, str] | None = None) -> HTMLResponse:
    """Return an HTML response."""
    return HTMLResponse(content, status_code, headers=headers)


def redirect(
    url: str | URL, status_code: int = 302, headers: typing.Mapping[str, str] | None = None
) -> RedirectResponse:
    """Redirect to a URL."""
    return RedirectResponse(url, status_code, headers=headers)


def redirect_to_route(
    request: Request,
    route_name: str,
    route_params: typing.Mapping[str, typing.Any] | None = None,
    status_code: int = 302,
    headers: typing.Mapping[str, str] | None = None,
) -> RedirectResponse:
    """Redirect to a named route."""
    url = request.url_for(route_name, **(route_params or {}))
    return RedirectResponse(url, status_code, headers=headers)


def json(
    content: typing.Any,
    status_code: int = 200,
    headers: typing.Mapping[str, str] | None = None,
    indent: int = 2,
    encoder_class: type[jsonlib.JSONEncoder] | None = None,
) -> JSONResponse:
    """Return a JSON response."""
    return JSONResponse(content, status_code, headers=headers, indent=indent, encoder_class=encoder_class)


def json_error(
    message: str,
    errors: dict[str, list[str]] | None = None,
    status_code: int = 200,
    headers: typing.Mapping[str, str] | None = None,
    media_type: str | None = None,
) -> JSONErrorResponse:
    return JSONErrorResponse(
        message=message,
        errors=errors,
        status_code=status_code,
        headers=headers,
        media_type=media_type,
    )


def send_file(
    path: str | os.PathLike[str],
    status_code: int = 200,
    headers: typing.Mapping[str, str] | None = None,
    media_type: str | None = None,
    filename: str | None = None,
    stat_result: os.stat_result | None = None,
    content_disposition_type: str = "attachment",
    inline: bool = False,
) -> FileResponse:
    """Return a response with a file."""
    if inline:
        content_disposition_type = "inline"

    return FileResponse(
        path,
        status_code=status_code,
        headers=headers,
        media_type=media_type,
        filename=filename,
        stat_result=stat_result,
        content_disposition_type=content_disposition_type,
    )


def empty(headers: typing.Mapping[str, str] | None = None) -> EmptyResponse:
    """Return an empty response with status code 204."""
    return EmptyResponse(headers=headers)


def back(request: Request, status_code: int = 302, headers: typing.Mapping[str, str] | None = None) -> GoBackResponse:
    """Redirect to the previous page."""
    return GoBackResponse(request, status_code=status_code, headers=headers)


def stream(
    content: ContentStream,
    status_code: int = 200,
    headers: typing.Mapping[str, str] | None = None,
    media_type: str = "application/octet-stream",
) -> StreamingResponse:
    return StreamingResponse(
        content=content,
        status_code=status_code,
        headers=headers,
        media_type=media_type,
    )


def template(
    request: Request,
    template_name: str,
    context: dict[str, typing.Any] | None = None,
    status_code: int = 200,
    headers: typing.Mapping[str, str] | None = None,
    media_type: str = "text/html",
) -> Response:
    """Render a template to a response."""
    return Templates.of(request.app).render_to_response(
        request, template_name, context, status_code=status_code, headers=headers, media_type=media_type
    )


def template_block(
    request: Request,
    template_name: str,
    block: str,
    context: dict[str, typing.Any] | None = None,
    status_code: int = 200,
    headers: typing.Mapping[str, str] | None = None,
    media_type: str = "text/html",
) -> Response:
    """Render a block from a template to a response."""
    content = Templates.of(request.app).render_block(template_name, block, context)
    return Response(content, status_code=status_code, headers=headers, media_type=media_type)


def template_macro(
    request: Request,
    template_name: str,
    macro_name: str,
    context: typing.Mapping[str, typing.Any] | None = None,
    status_code: int = 200,
    headers: typing.Mapping[str, str] | None = None,
    media_type: str = "text/html",
) -> Response:
    """Render a block from a template to a response."""
    content = Templates.of(request.app).render_macro(template_name, macro_name, context)
    return Response(content, status_code=status_code, headers=headers, media_type=media_type)



from __future__ import annotations

import functools
import inspect
import typing

from starlette.concurrency import run_in_threadpool
from starlette.datastructures import URL
from starlette.middleware import Middleware as ASGIMiddleware
from starlette.requests import HTTPConnection, Request
from starlette.responses import Response
from starlette.routing import BaseRoute, Mount, Route as Route_, WebSocketRoute as WebSocketRoute_, compile_path
from starlette.websockets import WebSocket

from kupala.asgi.request_limit import RequestBodyLimitMiddleware
from kupala.asgi.request_timeout import RequestTimeoutMiddleware
from kupala.dependencies import create_dependency_specs, Dependencies, VariableResolver

AsyncViewCallable = typing.Callable[..., typing.Awaitable[Response]]
SyncViewCallable = typing.Callable[..., Response]
AnyViewCallable = AsyncViewCallable | SyncViewCallable
WebSocketViewCallable = typing.Callable[[WebSocket], typing.Awaitable[None]]
HttpMethod = str

_PS = typing.ParamSpec("_PS")

type CallNext = typing.Callable[[Request], typing.Awaitable[Response]]
type Middleware = typing.Callable[[Request, CallNext], typing.Awaitable[Response]]


def unwrap_callable(fn: AnyViewCallable) -> AnyViewCallable:
    return fn if not hasattr(fn, "__wrapped__") else unwrap_callable(fn.__wrapped__)


def unwrap_websocket_callable(
    fn: typing.Callable[..., typing.Awaitable[None]],
) -> typing.Callable[..., typing.Awaitable[None]]:
    callback = fn if not hasattr(fn, "__wrapped__") else unwrap_callable(fn.__wrapped__)
    return typing.cast(typing.Callable[..., typing.Awaitable[None]], callback)


def wrap_with_middleware(fn: AsyncViewCallable, middleware: typing.Sequence[Middleware]) -> AsyncViewCallable:
    async def call_next(request_: Request) -> Response:
        return await fn(request_)

    async def middleware_adapter(middleware: Middleware, call_next_: CallNext, request_: Request) -> Response:
        return await middleware(request_, call_next_)

    for middleware_ in reversed(middleware):
        call_next = functools.partial(middleware_adapter, middleware_, call_next)
    return call_next


def get_view_metadata(view_callable: AnyViewCallable) -> dict[str, typing.Any]:
    return getattr(view_callable, "__metadata__", {})


class Route(Route_):
    def __init__(
        self,
        path: str,
        endpoint: typing.Callable[..., typing.Any],
        *,
        methods: list[str] | None = None,
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: typing.Sequence[Middleware] = (),
        metadata: dict[str, typing.Any] | None = None,
        max_request_size: int | None = None,
        request_timeout: float | None = None,
    ) -> None:
        setattr(endpoint, "__metadata__", metadata or {})
        self.middleware = [*middleware]

        super().__init__(
            path,
            name=name,
            methods=methods,
            include_in_schema=include_in_schema,
            endpoint=self._build_endpoint(endpoint),
            middleware=[
                ASGIMiddleware(RequestBodyLimitMiddleware, max_body_size=max_request_size),
                ASGIMiddleware(RequestTimeoutMiddleware, timeout=request_timeout),
            ],
        )

    def _build_endpoint(self, view_callable: typing.Callable[..., typing.Any]) -> AsyncViewCallable:
        # find the original view callable in order to parse the dependencies
        actual_view_callable = unwrap_callable(view_callable)
        dependencies = create_dependency_specs(actual_view_callable)

        async def endpoint(request: Request) -> Response:
            async def wrapped_endpoint(request: Request) -> Response:
                nonlocal view_callable
                resolver = Dependencies(
                    parent=request.app.dependencies,
                    dependencies={
                        Request: VariableResolver(request),
                        HTTPConnection: VariableResolver(request),
                    },
                )
                if inspect.iscoroutinefunction(view_callable) is False:
                    view_callable = functools.partial(run_in_threadpool, view_callable)

                return await resolver.call_with(view_callable, dependencies)

            wrapped = wrap_with_middleware(wrapped_endpoint, [*request.app.middleware, *self.middleware])
            return await wrapped(request)

        return endpoint


class WebSocketRoute(WebSocketRoute_):
    def __init__(
        self,
        path: str,
        endpoint: typing.Callable[..., typing.Any],
        *,
        name: str | None = None,
        metadata: dict[str, typing.Any] | None = None,
    ) -> None:
        self.metadata = metadata or {}
        super().__init__(
            path=path,
            endpoint=self._build_endpoint(endpoint),
            name=name,
        )

    def _build_endpoint(self, view_callable: typing.Callable[..., typing.Any]) -> AsyncViewCallable:
        # find the original view callable in order to parse the dependencies
        actual_view_callable = unwrap_websocket_callable(view_callable)
        dependencies = create_dependency_specs(actual_view_callable)

        async def endpoint(websocket: WebSocket) -> Response:
            nonlocal view_callable
            resolver = Dependencies(
                parent=websocket.app.dependencies,
                dependencies={
                    WebSocket: VariableResolver(websocket),
                    HTTPConnection: VariableResolver(websocket),
                },
            )

            return await resolver.call_with(view_callable, dependencies)

        return endpoint


class RouteGroup(typing.Sequence[BaseRoute]):
    def __init__(
        self,
        prefix: str | None = None,
        middleware: typing.Sequence[Middleware] = (),
        children: typing.Sequence[RouteGroup | BaseRoute] | None = None,
    ) -> None:
        if prefix and prefix.endswith("/"):
            raise ValueError("Prefix must not end with a slash.")

        self.prefix = prefix or ""
        self.routes: list[BaseRoute] = []
        self._group_middleware = list(middleware or [])

        for child in children or []:
            if isinstance(child, RouteGroup):
                self.merge(child)
            else:
                self.merge([child])

    def merge(self, group: typing.Sequence[BaseRoute]) -> None:
        for route in group:
            if self.prefix and isinstance(route, Route | Mount):
                route.path = self.prefix + route.path
                route.path_regex, route.path_format, route.param_convertors = compile_path(route.path)
            self.routes.append(route)

    def add(
        self,
        path: str,
        *,
        methods: list[HttpMethod] | None = None,
        name: str | None = None,
        middleware: typing.Sequence[Middleware] = (),
        metadata: dict[str, typing.Any] | None = None,
    ) -> typing.Callable[[AnyViewCallable], AnyViewCallable]:
        if not path.startswith("/"):
            raise ValueError("Path must start with a slash.")

        path = self.prefix + path if self.prefix else path

        def decorator(view_callable: AnyViewCallable) -> AnyViewCallable:
            self.routes.append(
                Route(
                    path,
                    view_callable,
                    name=name,
                    methods=methods,
                    metadata=metadata,
                    middleware=(*self._group_middleware, *middleware),
                )
            )
            return view_callable

        return decorator

    def get(
        self, path: str, *, name: str | None = None, middleware: typing.Sequence[Middleware] = ()
    ) -> typing.Callable[[AnyViewCallable], AnyViewCallable]:
        return self.add(path, methods=["GET"], name=name, middleware=middleware)

    def post(
        self, path: str, *, name: str | None = None, middleware: typing.Sequence[Middleware] = ()
    ) -> typing.Callable[[AnyViewCallable], AnyViewCallable]:
        return self.add(path, methods=["POST"], name=name, middleware=middleware)

    def get_or_post(
        self, path: str, *, name: str | None = None, middleware: typing.Sequence[Middleware] = ()
    ) -> typing.Callable[[AnyViewCallable], AnyViewCallable]:
        return self.add(path, methods=["GET", "POST"], name=name, middleware=middleware)

    def put(
        self, path: str, *, name: str | None = None, middleware: typing.Sequence[Middleware] = ()
    ) -> typing.Callable[[AnyViewCallable], AnyViewCallable]:
        return self.add(path, methods=["PUT"], name=name, middleware=middleware)

    def patch(
        self, path: str, *, name: str | None = None, middleware: typing.Sequence[Middleware] = ()
    ) -> typing.Callable[[AnyViewCallable], AnyViewCallable]:
        return self.add(path, methods=["PATCH"], name=name, middleware=middleware)

    def delete(
        self, path: str, *, name: str | None = None, middleware: typing.Sequence[Middleware] = ()
    ) -> typing.Callable[[AnyViewCallable], AnyViewCallable]:
        return self.add(path, methods=["DELETE"], name=name, middleware=middleware)

    def websocket(
        self, path: str, *, name: str | None = None
    ) -> typing.Callable[[typing.Callable[_PS, typing.Awaitable[None]]], typing.Callable[_PS, typing.Awaitable[None]]]:
        path = self.prefix.removesuffix("/") + path if self.prefix else path

        def decorator(
            view_callable: typing.Callable[_PS, typing.Awaitable[None]],
        ) -> typing.Callable[_PS, typing.Awaitable[None]]:
            self.routes.append(
                WebSocketRoute(
                    path,
                    view_callable,
                    name=name,
                )
            )
            return view_callable

        return decorator

    def __iter__(self) -> typing.Iterator[BaseRoute]:
        return iter(self.routes)

    def __len__(self) -> int:
        return len(self.routes)

    def __repr__(self) -> str:
        routes_count = len(self.routes)
        noun = "route" if routes_count == 1 else "routes"
        return f"<{self.__class__.__name__}: {routes_count} {noun}>"

    @typing.overload
    def __getitem__(self, index: int) -> BaseRoute:  # pragma: no cover
        ...

    @typing.overload
    def __getitem__(self, index: slice) -> typing.Sequence[BaseRoute]:  # pragma: no cover
        ...

    def __getitem__(self, index: int | slice) -> BaseRoute | typing.Sequence[BaseRoute]:
        return self.routes[index]


def url_matches(request: Request, pattern: URL | str) -> bool:
    """Return True if request URL matches URL."""
    value = URL(str(pattern))
    return request.url.path.removesuffix("/").startswith(value.path.removesuffix("/"))


def route_matches(request: Request, pathname: str, path_params: dict[str, typing.Any] | None = None) -> bool:
    url = request.url_for(pathname, **(path_params or {}))
    return url_matches(request, url)


def abs_url_for(request: Request, name: str, **path_params: typing.Any) -> URL:
    """Return absolute URL for route."""
    return request.url_for(name, **path_params)



import typing

from alembic.autogenerate.api import AutogenContext


class RendersMigrationType:
    """Renders proper migrations for custom SQLAlchemy types."""

    def render_item(self, _type: typing.Any, _obj: typing.Any, autogen_context: AutogenContext) -> str:
        autogen_context.imports.add(self.get_import_name())
        return self.__class__.__name__

    def get_import_name(self) -> str:
        module_name = self.__class__.__module__
        class_name = self.__class__.__name__
        return f"from {module_name} import {class_name}"



import contextlib
import typing
from unittest import mock

import click
import pytest
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from starlette.testclient import TestClient
from starlette.types import Scope

from kupala.applications import Kupala


class TestApp:
    async def test_sets_current_app(self) -> None:
        app = Kupala(secret_key="key!")
        async with app.initialize():
            assert Kupala.current() is app

        with pytest.raises(RuntimeError):
            Kupala.current()

    async def test_initializers(self) -> None:
        value = 0

        @contextlib.asynccontextmanager
        async def initializer(app: Kupala) -> typing.AsyncGenerator[None, None]:
            nonlocal value
            value = 1
            yield None
            value = 0

        app = Kupala(secret_key="key!", initializers=[initializer])
        async with app.initialize():
            assert value == 1
        assert value == 0

    def test_extensions(self) -> None:
        configure_callback = mock.MagicMock()
        extension = mock.MagicMock(configure=configure_callback)
        app = Kupala(secret_key="key!", extensions=[extension])
        configure_callback.assert_called_once_with(app)


class TestASGIApp:
    async def test_injects_app_into_scope(self) -> None:
        app = Kupala(secret_key="key!")
        scope: Scope = {"type": "http", "path": "/", "method": "GET", "headers": []}
        await app(scope, mock.AsyncMock(), mock.AsyncMock())
        assert scope["app"] is app


class TestCommandLineApp:
    def test_initializers(self) -> None:
        stack: list[int] = []

        @contextlib.asynccontextmanager
        async def initializer(app: Kupala) -> typing.AsyncGenerator[None, None]:
            stack.append(1)
            yield None
            stack.append(2)

        app = Kupala(secret_key="key!", initializers=[initializer])
        app.run_cli()
        assert stack == [1, 2]

    def test_calls_sync_command(self) -> None:
        spy = mock.MagicMock()

        @click.command("call-me")
        def command() -> None:
            spy()

        app = Kupala(secret_key="key!", commands=[command])
        app.run_cli("call-me")
        spy.assert_called_once()

    def test_calls_async_command(self) -> None:
        spy = mock.MagicMock()

        @click.command("call-me")
        async def command() -> None:
            spy()

        app = Kupala(secret_key="key!", commands=[command])
        app.run_cli("call-me")
        spy.assert_called_once()

    def test_unknown_command(self) -> None:
        spy = mock.MagicMock()

        app = Kupala(secret_key="key!")
        with pytest.raises(SystemExit):
            app.run_cli("call-me")
            spy.assert_called_once()

    def test_injects_app_as_obj(self) -> None:
        spy = mock.MagicMock()

        @click.command("call-me")
        @click.pass_obj
        def command(obj: Kupala) -> None:
            spy(obj)

        app = Kupala(secret_key="key!", commands=[command])
        app.run_cli("call-me")
        spy.assert_called_once_with(app)


class TestTrustedHosts:
    def test_trusted_hosts(self) -> None:
        app = Kupala(
            secret_key="key!",
            trusted_hosts=["example.com", "*.example.com"],
            routes=[
                Route("/", PlainTextResponse("OK")),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/", headers={"host": "example.com"})
            assert response.status_code == 200

            response = client.get("/", headers={"host": "sub.example.com"})
            assert response.status_code == 200

            response = client.get("/", headers={"host": "another.org"})
            assert response.status_code == 400


class TestCompressResponse:
    def test_compression_on(self) -> None:
        app = Kupala(
            secret_key="key!",
            compress_response=True,
            routes=[
                Route("/", PlainTextResponse("0" * 512)),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.headers["content-encoding"] == "gzip"

    def test_compression_off(self) -> None:
        app = Kupala(
            secret_key="key!",
            compress_response=False,
            routes=[
                Route("/", PlainTextResponse("0" * 512)),
            ],
        )
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 200
            assert "content-encoding" not in response.headers



from __future__ import annotations

import contextlib
import dataclasses
import datetime
import enum
import functools
import os
import time
import typing

import anyio
from async_storages import FileStorage, generate_file_path
from async_storages.backends.base import AsyncFileLike, AsyncReader
from async_storages.contrib.starlette import FileServer
from starlette.datastructures import UploadFile, URL
from starlette.requests import Request
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles as StarletteStaticFiles

from kupala import Kupala
from kupala.dependencies import VariableResolver
from kupala.templating import Templates

_boot_time = time.time()

KEY_STATIC_ROUTE_NAME = "staticfiles_route_name"
KEY_MEDIA_ROUTE_NAME = "mediafiles_route_name"


def static_url(request: Request, path: str) -> URL:
    """Return URL for static file.
    If path is absolute, return it as is."""
    if path.startswith(("http://", "https://")):
        return URL(path)

    path_name = request.app.state[KEY_STATIC_ROUTE_NAME]
    version = time.time() if request.app.debug else _boot_time
    return request.url_for(path_name, path=path).include_query_params(v=version)


def static_context_processor(request: Request) -> dict[str, typing.Any]:
    return {"static_url": functools.partial(static_url, request)}


class StaticFiles:
    def __init__(
        self,
        packages: list[str | tuple[str, str]] | None = None,
        directory: str | None = None,
        url_prefix: str = "/static",
        route_name: str = "static",
    ) -> None:
        self.url_prefix = url_prefix
        self.route_name = route_name
        self.packages = packages
        self.directory = directory

    def configure(self, app: Kupala) -> None:
        app.routes.append(
            Mount(
                path=self.url_prefix,
                name=self.route_name,
                app=StarletteStaticFiles(
                    packages=self.packages,
                    directory=self.directory,
                ),
            ),
        )
        app.state[KEY_STATIC_ROUTE_NAME] = self.route_name
        with contextlib.suppress(KeyError):
            templates: Templates = app.state[Templates]
            templates.context_processors.append(static_context_processor)


class StorageType(enum.StrEnum):
    S3 = "s3"
    LOCAL = "local"
    MEMORY = "memory"


@dataclasses.dataclass(frozen=True)
class LocalConfig:
    directory: str
    url_prefix: str
    make_dirs: bool = True
    make_dirs_exists_ok: bool = True
    make_dirs_permissions: int = 0o777


@dataclasses.dataclass(frozen=True)
class S3Config:
    bucket: str
    access_key: str
    secret_key: str
    region: str
    endpoint: str | None = None
    signed_link_ttl: datetime.timedelta = datetime.timedelta(hours=1)


@dataclasses.dataclass(frozen=True)
class MemoryConfig:
    pass


type StorageConfig = LocalConfig | S3Config | MemoryConfig


def media_url(request: Request, path: str) -> URL:
    """Return URL for static file.
    If path is absolute, return it as is."""
    if path.startswith(("http://", "https://")):
        return URL(path)

    path_name = request.app.state[KEY_MEDIA_ROUTE_NAME]
    return request.url_for(path_name, path=path)


def media_context_processor(request: Request) -> dict[str, typing.Any]:
    return {"media_url": functools.partial(static_url, request)}


class Files:
    def __init__(
        self,
        default: str,
        storages: dict[str, FileStorage],
        url_prefix: str = "/media",
        route_name: str = "media",
    ) -> None:
        assert default in storages, f"Default storage '{default}' is not configured."

        self._storages = storages or {}
        self._default = default
        self.url_prefix = url_prefix
        self.route_name = route_name

    def get_storage(self, name: str) -> FileStorage:
        return self._storages[name]

    async def exists(self, file_name: str | os.PathLike[typing.AnyStr], *, storage_name: str | None = None) -> bool:
        storage = self.get_storage(storage_name or self._default)
        return await storage.exists(file_name)

    async def open(
        self, file_name: str | os.PathLike[typing.AnyStr], *, storage_name: str | None = None
    ) -> AsyncFileLike:
        return await self.get_storage(storage_name or self._default).open(file_name)

    async def readlines(
        self, file_name: str | os.PathLike[typing.AnyStr], storage_name: str | None = None
    ) -> typing.AsyncIterable[bytes]:
        return await self.get_storage(storage_name or self._default).iterator(str(file_name))

    async def write(
        self,
        file_name: str | os.PathLike[typing.AnyStr],
        data: bytes | AsyncReader | typing.BinaryIO,
        *,
        storage_name: str | None = None,
    ) -> None:
        return await self.get_storage(storage_name or self._default).write(file_name, data)

    async def upload(
        self,
        upload_file: UploadFile,
        destination: str,
        *,
        extra_tokens: typing.Mapping[str, typing.Any] | None = None,
        storage_name: str | None = None,
    ) -> str:
        assert upload_file.filename, "Filename is required"
        storage = self.get_storage(storage_name or self._default)
        file_name = generate_file_path(upload_file.filename, destination, extra_tokens=extra_tokens or {})
        await storage.write(file_name, upload_file)
        return file_name

    async def upload_many(
        self,
        destination: str,
        upload_files: typing.Sequence[UploadFile],
        extra_tokens: typing.Mapping[str, typing.Any] | None = None,
    ) -> list[str]:
        file_names: list[str] = []
        extra_tokens = extra_tokens or {}

        async def worker(file: UploadFile) -> None:
            assert file.filename, "Filename is required"
            file_path = await self.upload(file, destination, extra_tokens=extra_tokens)
            file_names.append(file_path)

        async with anyio.create_task_group() as tg:
            for file in upload_files:
                tg.start_soon(worker, file)
        return file_names

    async def delete(
        self,
        file_name: str | os.PathLike[typing.AnyStr],
        *,
        storage_name: str | None = None,
    ) -> None:
        storage = self.get_storage(storage_name or self._default)
        await storage.delete(file_name)

    async def delete_many(
        self,
        file_names: typing.Sequence[str | os.PathLike[typing.AnyStr]],
        *,
        storage_name: str | None = None,
    ) -> None:
        async with anyio.create_task_group() as tg:
            for file_name in file_names:
                tg.start_soon(functools.partial(self.delete, storage_name=storage_name), file_name)

    def configure(self, app: Kupala) -> None:
        app.routes.append(
            Mount(
                path=self.url_prefix,
                name=self.route_name,
                app=FileServer(self._storages[self._default], as_attachment=True, redirect_status=301),
            ),
        )
        app.state[KEY_MEDIA_ROUTE_NAME] = self.route_name
        app.dependencies.registry[type(self)] = VariableResolver(self)
        app.state[self.__class__] = self
        with contextlib.suppress(KeyError):
            templates: Templates = app.state[Templates]
            templates.context_processors.append(media_context_processor)

    @classmethod
    def of(cls, app: Kupala) -> typing.Self:
        return typing.cast(typing.Self, app.state[cls])



from kupala.choices import IntegerChoices, TextChoices


def test_text_choices_without_label() -> None:
    class Fruits(TextChoices):
        APPLE = "apple"
        PEAR = "pear"
        BANANA = "banana"

    assert Fruits.APPLE == Fruits("apple")
    assert str(Fruits.APPLE) == "apple"
    assert Fruits.choices == (
        ("apple", "Apple"),
        ("pear", "Pear"),
        ("banana", "Banana"),
    )
    assert Fruits.labels == ("Apple", "Pear", "Banana")
    assert Fruits.values == ("apple", "pear", "banana")
    assert str(Fruits.APPLE) == "apple"


def test_text_choices_with_label() -> None:
    class Fruits(TextChoices):
        APPLE = ("apple", "Apple")
        PEAR = ("pear", "Pear")
        BANANA = ("banana", "Banana")

    assert Fruits.APPLE == "apple"
    assert Fruits.choices == (
        ("apple", "Apple"),
        ("pear", "Pear"),
        ("banana", "Banana"),
    )
    assert Fruits.labels == ("Apple", "Pear", "Banana")
    assert Fruits.values == ("apple", "pear", "banana")


def test_integer_choices_without_label() -> None:
    class Fruits(IntegerChoices):
        APPLE = 1, "Apple"
        PEAR = 2, "Pear"
        BANANA = 3, "Banana"

    assert Fruits.APPLE == Fruits(1)
    assert int(Fruits.APPLE) == 1
    assert Fruits.choices == (
        (1, "Apple"),
        (2, "Pear"),
        (3, "Banana"),
    )
    assert Fruits.labels == ("Apple", "Pear", "Banana")
    assert Fruits.values == (1, 2, 3)


def test_integer_choices_with_label() -> None:
    class Fruits(IntegerChoices):
        APPLE = 1
        PEAR = 2
        BANANA = 3

    assert Fruits.APPLE == 1
    assert Fruits.choices == (
        (1, "Apple"),
        (2, "Pear"),
        (3, "Banana"),
    )
    assert Fruits.labels == ("Apple", "Pear", "Banana")
    assert Fruits.values == (1, 2, 3)



import pytest
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.testclient import TestClient
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket

from kupala.asgi.request_limit import RequestBodyLimitMiddleware


async def app(scope: Scope, receive: Receive, send: Send) -> None:
    request = Request(scope, receive, send)
    await request.form()
    await PlainTextResponse("")(scope, receive, send)


@pytest.mark.asyncio
async def test_request_limit_middleware_limits() -> None:
    """When request body is larger than a configured limit then 413 response to be returned."""

    client = TestClient((RequestBodyLimitMiddleware(app, max_body_size=5)))
    assert client.post("/", data={"content": ""}).status_code == 413  # 8 bytes, message + "=" sign


@pytest.mark.asyncio
async def test_request_limit_middleware_allows() -> None:
    client = TestClient((RequestBodyLimitMiddleware(app, max_body_size=10)))
    assert client.post("/", data={"content": ""}).status_code == 200  # 8 bytes, message + "=" sign


@pytest.mark.asyncio
async def test_request_limit_middleware_ignores_websockets() -> None:
    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        ws = WebSocket(scope, receive, send)
        await ws.accept()
        await ws.receive_text()

    client = TestClient((RequestBodyLimitMiddleware(app, max_body_size=1)))
    with client.websocket_connect("/") as session:
        session.send_text("content")



import typing

from sqlalchemy.ext.asyncio import AsyncSession

from kupala.dependencies import RequestResolver

type DbSession = typing.Annotated[AsyncSession, RequestResolver(lambda r: r.state.dbsession)]



from __future__ import annotations

import abc
import contextlib
import dataclasses
import inspect
import types
import typing

from starlette.concurrency import run_in_threadpool
from starlette.requests import HTTPConnection


class DependencyError(Exception): ...


class DependencyNotFoundError(Exception): ...


class DependencyRequiresValueError(Exception): ...


@dataclasses.dataclass
class Argument:
    param_name: str
    param_type: type | typing.TypeAliasType
    default: typing.Any
    annotation: typing.Any
    nullable: bool


@dataclasses.dataclass
class ResolveContext:
    dependencies: Dependencies
    sync_stack: contextlib.ExitStack = dataclasses.field(default_factory=contextlib.ExitStack)
    async_stack: contextlib.AsyncExitStack = dataclasses.field(default_factory=contextlib.AsyncExitStack)
    cache: dict[int, typing.Any] = dataclasses.field(default_factory=dict)

    async def resolve(self, dependency: typing.Hashable | DependencyResolver) -> typing.Any:
        return await self.dependencies.resolve(self, dependency)

    async def __aenter__(self) -> typing.Self:
        self.sync_stack.__enter__()
        await self.async_stack.__aenter__()
        return self

    async def __aexit__(self, *exc_type: typing.Any) -> None:
        await self.async_stack.__aexit__(*exc_type)
        self.sync_stack.__exit__(*exc_type)


class DependencyResolver(abc.ABC):  # pragma: no cover
    @abc.abstractmethod
    async def resolve(self, context: ResolveContext) -> typing.Any:
        raise NotImplementedError


type DependencyScope = typing.Literal["transient", "cached", "scoped"]


class RegistryResolver(DependencyResolver):
    def __init__(self, klass: typing.Hashable) -> None:
        self.klass = klass

    async def resolve(self, context: ResolveContext) -> typing.Any:
        return await context.resolve(self.klass)


class FactoryResolver(DependencyResolver):
    """Dependency resolver that resolves dependencies from factories."""

    def __init__(self, factory: typing.Callable[..., typing.Any], *, scope: DependencyScope = "transient") -> None:
        self._scope = scope
        self._factory = factory
        self._dependencies = create_dependency_specs(factory)
        self._is_async = inspect.iscoroutinefunction(factory)
        self._value: typing.Any = None

    async def resolve(self, context: ResolveContext) -> typing.Any:
        if self._scope == "cached" and self._value is not None:
            return self._value

        if self._scope == "scoped":
            if value := context.cache.get(id(context)):
                return value

        resolver = Dependencies(parent=context.dependencies)
        parent_argument = await context.resolve(Argument)

        # replace Argument injections with parent one
        # it does not make much sense to know the argument name of factory
        # we are interested in the value of the dependant function argument info
        self_dependencies = {
            k: Injection(
                argument=v.argument,
                resolver=VariableResolver(parent_argument),
            )
            if v.argument.param_type == Argument
            else v
            for k, v in self._dependencies.items()
        }
        dependencies = await resolver.solve_parameters(context, self_dependencies)
        value = await self._resolve_function(dependencies)
        if isinstance(value, contextlib.AbstractContextManager):
            value = context.sync_stack.enter_context(value)
        elif isinstance(value, contextlib.AbstractAsyncContextManager):
            value = await context.async_stack.enter_async_context(value)
        else:
            value = value

        if self._scope == "scoped":
            context.cache[id(context)] = value

        if self._scope == "cached":
            self._value = value

        return value

    async def _resolve_function(self, dependencies: dict[str, typing.Any]) -> typing.Any:
        return await self._factory(**dependencies) if self._is_async else self._factory(**dependencies)


class VariableResolver(DependencyResolver):
    """Simple resolver that returns the same value for all dependencies."""

    __match_args__ = ("_value",)

    def __init__(self, value: typing.Any) -> None:
        self._value = value

    async def resolve(self, context: ResolveContext) -> typing.Any:
        return self._value


class RequestResolver(DependencyResolver):
    """Helper resolver that uses request state to return dependency values.
    It accepts a callable that receives HTTPConnection (like Request or WebSocket) and returns a value.

    Note: this resolver should be used in request context only.
    """

    def __init__(
        self,
        fn: typing.Callable[[HTTPConnection, Argument], typing.Any] | typing.Callable[[HTTPConnection], typing.Any],
    ) -> None:
        signature = inspect.signature(fn)
        self._takes_argument = len(signature.parameters) == 2
        self._fn = fn

    async def resolve(self, context: ResolveContext) -> typing.Any:
        args = [await context.resolve(HTTPConnection)]
        if self._takes_argument:
            args.append(await context.resolve(Argument))
        return self._fn(*args)


@dataclasses.dataclass
class Injection:
    argument: Argument
    resolver: DependencyResolver


def create_dependency_from_parameter(parameter: inspect.Parameter) -> Injection:
    annotation: typing.Any = parameter.annotation

    # test if annotation is a type alias
    # reconstruct the annotation with the original type
    # example: type FromJSON = DependencyResolver()
    # example: type FromJSON[T] = typing.Annotated[T, DependencyResolver()]
    if hasattr(annotation, "__value__"):
        # test if it is a generic type alias
        alias_args = typing.get_args(annotation)
        annotation = annotation.__value__
        annotation_args = typing.get_args(annotation)

        # if this is a generic type alias, make sure that the value is typing.Annotated
        # other options are not supported
        if alias_args:
            if typing.get_origin(annotation) is not typing.Annotated:
                raise DependencyError(
                    "Only Annotated generic types are supported as dependencies. "
                    f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
                )
            if len(annotation_args) < 2:
                raise DependencyError(
                    "Annotated generic type should have at least two arguments: type and value. "
                    f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
                )

            annotation = typing.Annotated[alias_args[0], *annotation_args[1:]]

    origin = typing.get_origin(annotation)
    is_nullable = False
    param_type: type = parameter.annotation
    resolver: DependencyResolver = RegistryResolver(param_type)

    if parameter.kind is inspect.Parameter.POSITIONAL_ONLY:
        raise DependencyError(
            "Dependencies with positional-only parameters are not supported. "
            f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
        )

    if parameter.default is not inspect.Parameter.empty:
        raise DependencyError(
            "Dependencies with default values are not supported. "
            f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
        )

    if origin not in [None, typing.Annotated, types.UnionType]:
        raise DependencyError(
            "Only Union and Annotated generic types are supported as dependencies. "
            f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
        )

    # handle union types and nullable types
    if origin is types.UnionType:
        type_args = typing.get_args(parameter.annotation)

        # support union type with at most two arguments
        if len(type_args) > 2:
            raise DependencyError(
                "Union type with more than two arguments is not supported. "
                f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
            )

        # support only nullable unions
        # we can't resolve non-distinct union types
        if None not in type_args and types.NoneType not in type_args:
            raise DependencyError(
                "Union type without None is not supported. "
                f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
            )

        is_nullable = type(None) in type_args
        annotation = [arg for arg in typing.get_args(parameter.annotation) if arg is not None][0]
        origin = typing.get_origin(annotation)
        param_type = annotation

    if origin is typing.Annotated:
        type_args = typing.get_args(annotation)
        defined_param_type = type_args[0]
        if type(defined_param_type) is typing.TypeVar:
            defined_param_type = defined_param_type.__default__
            if defined_param_type is typing.NoDefault:
                raise DependencyError(
                    "TypeVar without default type is not supported. "
                    f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
                )
        if typing.get_origin(defined_param_type) is types.UnionType:
            defined_param_type_args = [arg for arg in typing.get_args(defined_param_type)]
            if types.NoneType not in defined_param_type_args:
                raise DependencyError(
                    "Union type without None is not supported. "
                    f"Parameter: {parameter.name}, annotation: {parameter.annotation}."
                )

            defined_param_type = defined_param_type_args[0]
            is_nullable = True

        param_type = defined_param_type

        match type_args:
            case (*_, DependencyResolver() as defined_resolver):
                resolver = defined_resolver

            case (*_, fn) if inspect.isfunction(fn) and fn.__name__ == "<lambda>":
                signature = inspect.signature(fn)
                if len(signature.parameters) == 0:
                    resolver = FactoryResolver(fn)
                elif len(signature.parameters) == 1:

                    def one_param_callback(request: HTTPConnection) -> typing.Any:
                        return fn(request)

                    resolver = RequestResolver(one_param_callback)
                elif len(signature.parameters) == 2:

                    def two_params_callback(request: HTTPConnection, spec: Argument) -> typing.Any:
                        return fn(request, spec)

                    resolver = RequestResolver(two_params_callback)
                else:
                    raise DependencyError(
                        "Lambda passed as dependency should accept zero, one, or two parameters: "
                        "(lambda: ...), (lambda request: ...), or (lambda request, spec: ...)."
                    )

            case (*_, fn) if inspect.isgeneratorfunction(fn) or inspect.isasyncgenfunction(fn):
                raise DependencyError(
                    "Generators are not supported as dependency factories. "
                    "Use context managers or async context managers instead. "
                    f"Parameter: {parameter.name}, resolver: {fn}."
                )

            case (*_, fn) if inspect.isfunction(fn):
                resolver = FactoryResolver(fn)

            case (*_, value):
                resolver = VariableResolver(value)

            case _:  # pragma: no cover, we never reach this line
                ...

    return Injection(
        resolver=resolver,
        argument=Argument(
            param_type=param_type,
            nullable=is_nullable,
            default=parameter.default,
            param_name=parameter.name,
            annotation=parameter.annotation,
        ),
    )


def create_dependency_specs(fn: typing.Callable[..., typing.Any]) -> dict[str, Injection]:
    signature = inspect.signature(fn, eval_str=True)
    return {parameter.name: create_dependency_from_parameter(parameter) for parameter in signature.parameters.values()}


_PS = typing.ParamSpec("_PS")
_RT = typing.TypeVar("_RT")


class Dependencies:
    def __init__(
        self,
        dependencies: dict[typing.Hashable, DependencyResolver] | None = None,
        parent: Dependencies | None = None,
    ) -> None:
        self.parent = parent
        self.registry: dict[typing.Hashable, DependencyResolver] = dependencies or {}

    async def resolve(self, context: ResolveContext, dependency: DependencyResolver | typing.Hashable) -> typing.Any:
        if dependency in self.registry:
            return await self.registry[dependency].resolve(context)

        if typing.get_origin(dependency) is typing.Annotated:
            dependency = next((r for r in typing.get_args(dependency) if isinstance(r, DependencyResolver)), None)
            if not dependency:
                raise DependencyNotFoundError(f"Dependency {dependency} not found and is not a resolver.")

        if isinstance(dependency, DependencyResolver):
            return await dependency.resolve(context)

        if self.parent is not None:
            return await self.parent.resolve(context, dependency)

        raise DependencyNotFoundError(f"Dependency {dependency} not found.")

    async def solve_parameters(
        self, context: ResolveContext, dependencies: dict[str, Injection]
    ) -> dict[str, typing.Any]:
        solved_dependencies: dict[str, typing.Any] = {}
        for spec in dependencies.values():
            child_context = ResolveContext(
                sync_stack=context.sync_stack,
                async_stack=context.async_stack,
                dependencies=Dependencies(
                    parent=context.dependencies,
                    dependencies={Argument: VariableResolver(spec.argument)},
                ),
            )
            solved_dependencies[spec.argument.param_name] = await context.dependencies.resolve(
                child_context, spec.resolver
            )

        return solved_dependencies

    async def call_with(
        self,
        fn: typing.Callable[..., _RT] | typing.Callable[..., typing.Awaitable[_RT]],
        dependencies: dict[str, Injection],
    ) -> _RT:
        """Call a function with prepared dependencies."""
        context = ResolveContext(dependencies=self)

        async with context:
            solved_dependencies = await self.solve_parameters(context, dependencies)
            if inspect.iscoroutinefunction(fn):
                rv = await fn(**solved_dependencies)
            else:
                rv = await run_in_threadpool(fn, **solved_dependencies)
            return typing.cast(_RT, rv)


T = typing.TypeVar("T", default=str)

FromPath = typing.Annotated[
    T, RequestResolver(lambda request, attr: attr.param_type(request.path_params[attr.param_name]))
]



import datetime
from unittest import mock
import uuid

import pytest
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.authentication import UnauthenticatedUser
from starlette.requests import HTTPConnection, Request
from starlette.responses import Response
from starlette.testclient import TestClient

from kupala import Kupala
from kupala.authentication import (
    SESSION_HASH_KEY,
    SESSION_IDENTITY_KEY,
    Authentication,
    AuthenticationError,
    Authenticator,
    AuthenticatorBackend,
    BaseUser,
    ChainAuthenticator,
    CurrentUser,
    DatabaseUserLoader,
    HTTPBasicAuthenticator,
    Identity,
    SessionAuthenticator,
    make_session_auth_hash,
    update_session_auth_hash,
    user_registered,
)
from kupala.passwords import PasswordHasher
from kupala.routing import CallNext, RouteGroup
from tests.models import User


class MemoryUserLoader:
    def __init__(self, users: list[User]) -> None:
        self.users: dict[str, User] = {user.identity: user for user in users}

    async def __call__(self, conn: HTTPConnection, identity: str) -> BaseUser | None:
        return self.users.get(identity)


class _DummyAuthenticator(Authenticator):
    def __init__(self, user: BaseUser | None) -> None:
        self.user = user

    async def authenticate(self, conn: HTTPConnection) -> Identity | None:
        if self.user:
            return Identity(
                id=self.user.identity,
                user=self.user,
                scopes=[],
                authenticator=type(self),
            )
        return None


class TestSessionAuthenticator:
    async def test_authenticates(self) -> None:
        user = User(email="root")
        user_loader = MemoryUserLoader([user])

        conn = HTTPConnection(
            {
                "type": "http",
                "app": Kupala(secret_key="key!"),
                "session": {SESSION_IDENTITY_KEY: "root"},
            }
        )
        authenticator = SessionAuthenticator(user_loader=user_loader)
        identity = await authenticator.authenticate(conn)
        assert identity
        assert identity.id == "root"

    async def test_not_authenticated(self) -> None:
        user = User(email="root")
        user_loader = MemoryUserLoader([user])

        conn = HTTPConnection(
            {
                "type": "http",
                "app": Kupala(secret_key="key!"),
                "session": {},
            }
        )
        authenticator = SessionAuthenticator(user_loader=user_loader)
        assert not await authenticator.authenticate(conn)

    async def test_no_user(self) -> None:
        user = User(email="test")
        user_loader = MemoryUserLoader([user])

        conn = HTTPConnection(
            {
                "type": "http",
                "app": Kupala(secret_key="key!"),
                "session": {SESSION_IDENTITY_KEY: "root"},
            }
        )
        authenticator = SessionAuthenticator(user_loader=user_loader)
        assert not await authenticator.authenticate(conn)

    async def test_extracts_user_scopes(self) -> None:
        user = User(email="root", scopes=["admin"])
        user_loader = MemoryUserLoader([user])

        authenticator = SessionAuthenticator(user_loader=user_loader)
        conn = HTTPConnection(
            {
                "type": "http",
                "app": Kupala(secret_key="key!"),
                "session": {SESSION_IDENTITY_KEY: "root"},
            }
        )
        identity = await authenticator.authenticate(conn)
        assert identity
        assert identity.scopes == ["admin"]

    async def test_validates_session_hash(self) -> None:
        user = User(email="root", password="password")
        user_loader = MemoryUserLoader([user])
        user_loader = MemoryUserLoader([user])

        authenticator = SessionAuthenticator(user_loader=user_loader)
        conn = HTTPConnection(
            {
                "type": "http",
                "app": Kupala(secret_key="key!"),
                "session": {
                    SESSION_IDENTITY_KEY: "root",
                    SESSION_HASH_KEY: make_session_auth_hash(user, "key!"),
                },
            }
        )
        assert await authenticator.authenticate(conn)

        # simulate password change flow
        user.password = "new password"
        assert not await authenticator.authenticate(conn)
        update_session_auth_hash(conn, user, "key!")
        assert await authenticator.authenticate(conn)

    async def test_validates_invalid_session_hash(self) -> None:
        user = User(email="root", password="password")
        user_loader = MemoryUserLoader([user])

        authenticator = SessionAuthenticator(user_loader=user_loader)
        conn = HTTPConnection(
            {
                "type": "http",
                "app": Kupala(secret_key="key!"),
                "session": {
                    SESSION_IDENTITY_KEY: "root",
                    SESSION_HASH_KEY: "bad hash",
                },
            }
        )
        assert not await authenticator.authenticate(conn)

    async def test_update_session_hash(self) -> None:
        user = User(email="root", password="password")
        user_loader = MemoryUserLoader([user])

        authenticator = SessionAuthenticator(user_loader=user_loader)
        conn = HTTPConnection(
            {
                "type": "http",
                "app": Kupala(secret_key="key!"),
                "session": {
                    SESSION_IDENTITY_KEY: "root",
                    SESSION_HASH_KEY: "bad hash",
                },
            }
        )
        assert not await authenticator.authenticate(conn)

        update_session_auth_hash(conn, user, "key!")
        assert await authenticator.authenticate(conn)


class TestChainAuthenticator:
    async def test_authenticates(self) -> None:
        user = User(email="root")
        authenticator = ChainAuthenticator(
            _DummyAuthenticator(None),
            _DummyAuthenticator(user),
        )
        conn = HTTPConnection({"type": "http"})
        assert await authenticator.authenticate(conn)

    async def test_not_authenticates(self) -> None:
        backend = ChainAuthenticator(
            _DummyAuthenticator(None),
            _DummyAuthenticator(None),
        )
        conn = HTTPConnection({"type": "http"})
        assert not await backend.authenticate(conn)


class TestAuthenticatorBackend:
    async def test_authenticates(self) -> None:
        user = User(email="root")
        backend = AuthenticatorBackend(_DummyAuthenticator(user))
        result = await backend.authenticate(HTTPConnection({"type": "http"}))
        assert result
        credentials, authenticated_user = result
        assert credentials.scopes == []
        assert authenticated_user.identity == "root"

    async def test_not_authenticates(self) -> None:
        backend = AuthenticatorBackend(_DummyAuthenticator(None))
        assert not await backend.authenticate(HTTPConnection({"type": "http"}))


class TestHTTPBasicAuthenticator:
    async def test_authenticates_with_valid_credentials(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(
                averify=mock.AsyncMock(return_value=True),
            ),
        )

        user = User(email="root", password="hashed_password")
        user_loader = MemoryUserLoader([user])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [
                    (b"authorization", "Basic cm9vdDpjb3JyZWN0X3Bhc3N3b3Jk".encode("utf-8"))
                ],  # root:correct_password
            }
        )

        identity = await authenticator.authenticate(conn)

        assert identity is not None
        assert identity.id == "root"
        assert identity.user == user

    async def test_updates_session_auth_hash(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(
                averify=mock.AsyncMock(return_value=True),
            ),
        )

        user = User(email="root", password="hashed_password")
        user_loader = MemoryUserLoader([user])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [
                    (b"authorization", "Basic cm9vdDpjb3JyZWN0X3Bhc3N3b3Jk".encode("utf-8"))
                ],  # root:correct_password
                "session": {},
            }
        )

        await authenticator.authenticate(conn)
        assert SESSION_HASH_KEY in conn["session"]

    async def test_fails_with_invalid_password(self) -> None:
        app = Kupala(secret_key="key!")
        hasher = PasswordHasher()
        app.state[Authentication] = mock.MagicMock(
            passwords=hasher,
        )

        user = User(email="root", password=hasher.make("secret"))
        user_loader = MemoryUserLoader([user])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [(b"authorization", b"Basic cm9vdDpjb3JyZWN0X3Bhc3N3b3Jk")],  # root:correct_password
            }
        )

        assert not await authenticator.authenticate(conn)

    async def test_fails_with_nonexistent_user(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(spec=PasswordHasher),
        )

        user_loader = MemoryUserLoader([])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [(b"authorization", b"Basic cm9vdDpjb3JyZWN0X3Bhc3N3b3Jk")],  # root:correct_password
            }
        )

        assert not await authenticator.authenticate(conn)

    async def test_fails_with_no_authorization_header(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(spec=PasswordHasher),
        )

        user_loader = MemoryUserLoader([])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection({"type": "http", "app": app, "headers": []})

        assert not await authenticator.authenticate(conn)

    async def test_fails_with_non_basic_auth(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(spec=PasswordHasher),
        )

        user_loader = MemoryUserLoader([])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [(b"authorization", b"Key abc")],
            }
        )

        assert not await authenticator.authenticate(conn)

    async def test_raises_error_with_invalid_basic_header(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(spec=PasswordHasher),
        )

        user_loader = MemoryUserLoader([])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [(b"authorization", b"Basic")],
            }
        )

        with pytest.raises(AuthenticationError, match="Invalid basic header. No credentials provided."):
            assert not await authenticator.authenticate(conn)

    async def test_raises_error_with_invalid_basic_header_spaces(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(spec=PasswordHasher),
        )

        user_loader = MemoryUserLoader([])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [(b"authorization", b"Basic abc abc")],
            }
        )

        with pytest.raises(AuthenticationError, match="Invalid basic header. No spaces allowed."):
            assert not await authenticator.authenticate(conn)

    async def test_raises_error_with_invalid_base64_encoding(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(spec=PasswordHasher),
        )

        user_loader = MemoryUserLoader([])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [(b"authorization", b"Basic abc")],
            }
        )

        with pytest.raises(
            AuthenticationError, match="Invalid basic header. Credentials not correctly base64 encoded."
        ):
            assert not await authenticator.authenticate(conn)

    async def test_extracts_user_scopes(self) -> None:
        app = Kupala(secret_key="key!")
        app.state[Authentication] = mock.MagicMock(
            passwords=mock.AsyncMock(
                averify=mock.AsyncMock(return_value=True),
            ),
        )

        user = User(email="root", password="hashed_password", scopes=["admin"])
        user_loader = MemoryUserLoader([user])
        authenticator = HTTPBasicAuthenticator(user_loader)

        conn = HTTPConnection(
            {
                "type": "http",
                "app": app,
                "headers": [(b"authorization", b"Basic cm9vdDpjb3JyZWN0X3Bhc3N3b3Jk")],  # root:correct_password
            }
        )

        identity = await authenticator.authenticate(conn)
        assert identity
        assert identity.scopes == ["admin"]


class TestDependencies:
    def test_current_user(self) -> None:
        group = RouteGroup()

        @group.get("/")
        async def view(user: CurrentUser[User]) -> Response:
            return Response(user.identity)

        async def user_middleware(request: Request, call_next: CallNext) -> Response:
            request.scope["user"] = User(email="1")
            return await call_next(request)

        app = Kupala(secret_key="key!", routes=group, middleware=[user_middleware])
        with TestClient(app) as client:
            response = client.get("/")
            assert response.text == "1"

    def test_current_user_error(self) -> None:
        group = RouteGroup()

        @group.get("/")
        async def view(user: CurrentUser[User]) -> Response:
            return Response(user.identity)

        async def user_middleware(request: Request, call_next: CallNext) -> Response:
            request.scope["user"] = UnauthenticatedUser()
            return await call_next(request)

        app = Kupala(secret_key="key!", routes=group, middleware=[user_middleware])
        with TestClient(app) as client:
            response = client.get("/")
            assert response.status_code == 401


class TestSQLAlchemyUserLoader:
    async def test_loads_user_by_identity(self, dbsession: AsyncSession) -> None:
        loader = DatabaseUserLoader(User, lambda identity: User.id == identity)
        user = await loader(
            HTTPConnection(
                {
                    "type": "http",
                    "state": {"dbsession": dbsession},
                }
            ),
            "1",
        )
        assert user
        assert user.id == 1

    async def test_not_loads_user_by_identity(self, dbsession: AsyncSession) -> None:
        loader = DatabaseUserLoader(User, lambda identity: User.id == identity)
        user = await loader(
            HTTPConnection(
                {
                    "type": "http",
                    "state": {"dbsession": dbsession},
                }
            ),
            "-1",
        )
        assert not user


class TestBaseUser:
    def test_identity(self) -> None:
        email = uuid.uuid4().hex
        user = User(email=email)
        assert user.identity == email

    def test_display_name(self) -> None:
        user = User(email="1")
        assert user.display_name == "1"

    def test_is_authenticated(self) -> None:
        user = User(email="1")
        assert user.is_authenticated

    def test_is_active(self) -> None:
        user = User(email="1")
        assert user.is_active

    def test_is_not_active(self) -> None:
        user = User(email="1", deactivated_at=datetime.datetime.now(datetime.UTC))
        assert not user.is_active

    def test_get_password_hash(self) -> None:
        user = User(email="1", password="password")
        assert user.get_password_hash() == "password"

    def test_get_scopes(self) -> None:
        user = User(email=uuid.uuid4().hex, scopes=[])
        assert user.get_scopes() == []

    async def test_register(self, dbsession: AsyncSession) -> None:
        email = uuid.uuid4().hex
        spy = mock.AsyncMock()
        async with dbsession.begin_nested():
            user_registered.connect(spy)
            user = await User.register(dbsession, email=email, password_hash="password", name="name")
            await dbsession.commit()

        spy.assert_called_once_with(User, user=user)
        assert user.email == email



from __future__ import annotations

import contextlib
import contextvars
import functools
import inspect
import typing

import anyio
import click
from starlette.applications import Starlette
from starlette.datastructures import URLPath
from starlette.middleware import Middleware as ASGIMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.routing import BaseRoute
from starlette.types import ExceptionHandler, Receive, Scope, Send

from kupala.dependencies import Dependencies
from kupala.routing import Middleware


class Extension(typing.Protocol):
    def configure(self, app: Kupala) -> None: ...


type Initializer = typing.Callable[[Kupala], typing.AsyncContextManager[None]]


@contextlib.asynccontextmanager
async def _starlette_lifespan_adapter(_: Starlette, kupala: Kupala) -> typing.AsyncGenerator[None, None]:
    async with kupala.initialize():
        yield None


class Kupala:
    _current_app: contextvars.ContextVar[Kupala] = contextvars.ContextVar("kupala_app")

    def __init__(
        self,
        secret_key: str,
        *,
        debug: bool = False,
        middleware: typing.Sequence[Middleware] = (),
        asgi_middleware: typing.Sequence[ASGIMiddleware] = (),
        routes: typing.Sequence[BaseRoute] = (),
        initializers: typing.Sequence[Initializer] = (),
        extensions: typing.Sequence[Extension] = (),
        commands: typing.Sequence[click.Command] = (),
        dependencies: Dependencies | None = None,
        exception_handlers: typing.Mapping[typing.Any, ExceptionHandler] | None = None,
        max_body_size: int = 20 * 1024 * 1024,
        request_timeout: float = 30,
        trusted_hosts: typing.Sequence[str] = ("localhost", "127.0.0.1", "testserver"),
        compress_response: bool = False,
    ) -> None:
        self.debug = debug
        self.commands = list(commands)
        self.secret_key = secret_key
        self.middleware = middleware
        self.initializers = list(initializers)
        self.asgi_middleware = list(asgi_middleware)
        self.state: dict[typing.Any, typing.Any] = {}
        self.dependencies = dependencies or Dependencies()
        self.max_body_size = max_body_size
        self.request_timeout = request_timeout
        self.routes: list[BaseRoute] = list(routes)

        # setup trusted hosts
        if trusted_hosts:
            self.asgi_middleware.append(ASGIMiddleware(TrustedHostMiddleware, allowed_hosts=trusted_hosts))

        if compress_response:
            self.asgi_middleware.append(ASGIMiddleware(GZipMiddleware))

        for extension in extensions:
            extension.configure(self)

        self._app = Starlette(
            debug=debug,
            routes=self.routes,
            middleware=self.asgi_middleware,
            lifespan=functools.partial(_starlette_lifespan_adapter, kupala=self),
            exception_handlers=exception_handlers or {},
        )

    @classmethod
    def current(cls) -> Kupala:
        try:
            return cls._current_app.get()
        except LookupError:
            raise RuntimeError("No Kupala application is currently running.")

    def url_path_for(self, name: str, /, **path_params: typing.Any) -> URLPath:
        return self._app.router.url_path_for(name, **path_params)

    @contextlib.asynccontextmanager
    async def initialize(self) -> typing.AsyncGenerator[dict[str, typing.Any], None]:
        old_app = self._current_app.set(self)
        async with contextlib.AsyncExitStack() as stack:
            for initializer in self.initializers:
                await stack.enter_async_context(initializer(self))

            yield {}
            self._current_app.reset(old_app)

    def cli_plugin(self, app: click.Group) -> None:
        """Install this application as Kupala CLI plugin."""

        for command in self.commands:
            app.add_command(command)

    def run_cli(self, *args: str) -> None:
        """Run CLI application."""

        @click.group()
        @click.pass_context
        def cli(ctx: click.Context) -> None:
            ctx.obj = self

        self.cli_plugin(cli)

        async def main() -> None:
            async with self.initialize():
                try:
                    rv = cli(standalone_mode=False, args=args)
                    if inspect.iscoroutine(rv):
                        await rv
                except click.ClickException as exc:
                    click.secho("error: " + str(exc), err=True, fg="red")
                    raise SystemExit(exc.exit_code)

        anyio.run(main)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self

        if self._app.middleware_stack is None:
            self._app.middleware_stack = self._app.build_middleware_stack()

        await self._app.middleware_stack(scope, receive, send)



import os
from unittest import mock

import pytest
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession

from kupala.database.manager import Database, NoActiveSession, on_commit
from kupala.database.models import Base
from tests.models import User

POSTGRES_URL = os.environ.get("POSTGRES_URL", "postgresql+psycopg_async://postgres:postgres@127.0.0.1/test_kupala")


class TestDatabase:
    async def test_manager_initializes_engine(self) -> None:
        database = Database("sqlite+aiosqlite:///:memory:")
        async with database as engine:
            async with engine.begin() as conn:
                result = await conn.execute(sa.text("SELECT 1"))
                assert result.scalar() == 1

    async def test_creates_normal_session(self) -> None:
        database = Database("sqlite+aiosqlite:///:memory:")
        async with database as engine:
            async with engine.begin() as conn:
                await conn.run_sync(Base.metadata.create_all)

            async with database.session() as session:
                user = User(id=1, name="test", email="test", password="test")
                session.add(user)
                await session.commit()

            async with database.session() as session:
                result = await session.execute(sa.text("select count(*) from users"))
                assert result.scalar() == 1
                assert database.current_session is session

    async def test_retrieve_current_session(self) -> None:
        database = Database("sqlite+aiosqlite:///:memory:")
        async with database as engine:
            async with engine.begin() as conn:
                await conn.run_sync(Base.metadata.create_all)

            async with database.session() as parent_session:
                assert database.current_session is parent_session

                async with database.session() as child_session:
                    assert database.current_session is child_session

                assert database.current_session is parent_session

            with pytest.raises(NoActiveSession):
                assert database.current_session

    async def test_creates_rolling_back_session(self) -> None:
        database = Database(POSTGRES_URL)
        async with database as engine:
            async with engine.begin() as conn:
                await conn.run_sync(Base.metadata.drop_all)
                await conn.run_sync(Base.metadata.create_all)

            async with database.session(force_rollback=True) as session:
                user = User(id=1, name="test", email="test", password="test")
                session.add(user)
                await session.commit()

            async with database.session() as session:
                result = await session.execute(sa.select(User))
                users = result.scalars().all()
                assert len(users) == 0
                assert database.current_session is session


async def test_on_commit(dbsession: AsyncSession) -> None:
    spy = mock.MagicMock()
    spy2 = mock.MagicMock()
    on_commit(dbsession, spy)
    on_commit(dbsession, spy2)
    await dbsession.commit()
    spy.assert_called_once()
    spy2.assert_called_once()



"""SQLAlchemy column types and annotations."""

import datetime
import typing
import uuid

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import mapped_column

type JSONSerializable = str | int | float | bool | None | dict[str, JSONSerializable] | list[JSONSerializable]

IntPk = typing.Annotated[int, mapped_column(sa.BigInteger, primary_key=True, autoincrement=True)]
StrPk = typing.Annotated[str, mapped_column(sa.Text, primary_key=True)]
UUIDPk = typing.Annotated[uuid.UUID, mapped_column(sa.UUID(), primary_key=True, default=uuid.uuid4)]
JSONDict = typing.Annotated[dict[str, JSONSerializable], mapped_column(JSONB, default=dict, server_default="{}")]
JSONList = typing.Annotated[list[JSONSerializable], mapped_column(JSONB, default=list, server_default="[]")]

AutoCreatedAt = typing.Annotated[
    datetime.datetime,
    mapped_column(
        nullable=False,
        server_default=sa.func.now(),
        default=lambda: datetime.datetime.now(datetime.UTC),
    ),
]
AutoUpdatedAt = typing.Annotated[
    datetime.datetime,
    mapped_column(
        nullable=False,
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        onupdate=lambda: datetime.datetime.now(datetime.UTC),
        default=lambda: datetime.datetime.now(datetime.UTC),
    ),
]



from __future__ import annotations

import enum
import typing


T = typing.TypeVar("T")
Choices = typing.Iterable[tuple[T, str]]


class ChoicesMeta(enum.EnumMeta):
    _value_map: dict[typing.Any, str]

    def __new__(mcs, name: str, bases: tuple[typing.Any], attrs: typing.Any, **kwargs: typing.Any) -> ChoicesMeta:
        value_map = {}
        for key in attrs._member_names:
            member = attrs[key]
            match member:
                case list([value, label]) | tuple([value, label]):
                    value_map[value] = label
                    dict.__setitem__(attrs, key, value)
                case _:
                    value_map[member] = key.replace("_", " ").title()

        cls = super().__new__(mcs, name, bases, attrs, **kwargs)
        setattr(cls, "_value_map", value_map)
        return enum.unique(cls)  # type:ignore

    @property
    def labels(cls) -> typing.Iterable[str]:
        return tuple(cls._value_map.values())

    @property
    def values(cls) -> typing.Iterable[typing.Any]:
        return tuple(cls._value_map.keys())

    @property
    def choices(cls) -> Choices[typing.Any]:
        return tuple([tuple([value, label]) for value, label in cls._value_map.items()])


class BaseChoices(enum.Enum, metaclass=ChoicesMeta): ...


class IntegerChoices(int, BaseChoices):
    @typing.overload
    def __eq__(self, other: IntegerChoices) -> bool: ...

    @typing.overload
    def __eq__(self, other: int) -> bool: ...

    @typing.overload
    def __eq__(self, other: object) -> bool: ...

    def __eq__(self, other: object | int | IntegerChoices) -> bool:
        return self.value == other


class TextChoices(str, BaseChoices):
    def __str__(self) -> str:
        return str(self.value)

    @typing.overload
    def __eq__(self, other: TextChoices) -> bool: ...

    @typing.overload
    def __eq__(self, other: str) -> bool: ...

    @typing.overload
    def __eq__(self, other: object) -> bool: ...

    def __eq__(self, other: object | str | TextChoices) -> bool:
        return self.value == other



from __future__ import annotations

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column, relationship

from kupala.authentication import BaseUser
from kupala.database.models import Base


class User(BaseUser):
    __tablename__ = "users"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str]
    email: Mapped[str]
    password: Mapped[str] = mapped_column(default="__password_hash__")
    profile: Mapped[Profile] = relationship("Profile", back_populates="user")
    scopes: Mapped[list[str]] = mapped_column(sa.JSON, default=list)

    @property
    def identity(self) -> str:
        return self.email

    def get_scopes(self) -> list[str]:
        return self.scopes


class Profile(Base):
    __tablename__ = "profiles"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    user_id: Mapped[int] = mapped_column(sa.ForeignKey("users.id"))
    bio: Mapped[str]
    user: Mapped[User] = relationship("User", back_populates="profile")
